#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
"""
Module where the crypto library are defined.
"""

import abc
import logging
import sys

from ibmfl.crypto.crypto_exceptions import CryptoException, KeyManagerException
from ibmfl.model.model_update import ModelUpdate
from ibmfl.util.config import get_class_by_name

logger = logging.getLogger(__name__)

# the default precision setting.
PRECISION = 5


class Crypto(abc.ABC):
    """
    The base class for a crypto system implementation that will be exposed to
    other FL modules.
    """

    def __init__(self, config, **kwargs):
        self.cipher = None
        if not config:
            raise CryptoException("No crypto config is provided")
        if "precision" in config:
            if isinstance(config["precision"], int) and (config["precision"] > 0):
                self.precision = config["precision"]
            else:
                logger.warning(
                    "Provided precision is not in the correct format. "
                    "An accepted precision should be a positive integer. "
                    "Setting the precision to default precision: " + str(PRECISION)
                )
                self.precision = PRECISION
        else:
            self.precision = PRECISION

        self.idx = None
        logger.info("Initializing a key manager")
        if "key_manager" not in config:
            raise CryptoException("no key manager provided")
        key_mgr_config = config["key_manager"]
        try:
            key_mgr_cls_ref = get_class_by_name(key_mgr_config["path"], key_mgr_config["name"])
            self.key_manager = key_mgr_cls_ref(key_mgr_config["key_mgr_info"])
            self.idx = key_mgr_config["key_mgr_info"].get("idx", None)
        except KeyManagerException as ex:
            logger.exception(ex)

        self.initialize(config)

        self.ph = kwargs.get("proto_handler")

    @abc.abstractmethod
    def initialize(self, config):
        """
        Initialize a `KeyManager` object to generate keys for `Cipher` object;
        Initialize a 'Cipher' object

        :param config: A dictionary containing the provided crypto_keys,
            including public and private key(s), for encryption and decryption;
         etc., to initialize a cipher of a specific cryptosystem.
        :type config: `dict`
        :param kwargs: Dictionary of arguments required by a crypto system \
        implementation for encryption and decryption.
        :return: None
        """
        raise NotImplementedError

    def set_keys(self, keys):
        """
        Set the keys in use cases where the keys are not set during initialization
        of this object, e.g. for the keys generation and distribution protocol use cases.

        :param keys: The provided keys.
        :type keys: `dict`
        :return: None
        """
        self.cipher.set_keys(keys)
        return

    def encrypt(self, model_update, key="weights", **kwargs):
        """
        Method for the encryption operations FL party will perform
        during training. It currently calls the `encrypt_weights` method
        defined inside the Cipher class.

        :param model_update: ModelUpdate object to be encrypted.
        :type model_update: `ModelUpdate`
        :param key: model
        :type key: `str`
        :return: The resulting ciphertext of model update.
        :type: `ModelUpdate`
        """
        if not isinstance(model_update, ModelUpdate):
            raise CryptoException("not a ModelUpdate instance to be encrypted")

        model_weights = model_update.get(key)
        ct_model_update = ModelUpdate()
        if isinstance(model_weights, list):
            ct_model_update.add("ct_{}".format(key), self.cipher.encrypt_weights(model_weights))
        else:
            ct_model_update.add("ct_{}".format(key), self.cipher.encrypt_weights([model_weights]))
        return ct_model_update

    @abc.abstractmethod
    def decrypt(self, ct_model_update, key="ct_weights", **kwargs):
        """
        Abstract method to perform decryption or partial decryption operations
        FL party will perform during training.
        Decrypts the ciphertext included inside the provided `model_update`.
        The ciphertext is indicated by the dict_key.
        :param ct_model_update: A `ModelUpdate` object containing the provided \
        ciphertext to be decrypted. The ciphertext should be of \
        type `list` of `np.ndarray`.
        :type ct_model_update: `ModelUpdate`
        :param key: A dictionary key value indicating what encrypted \
        values inside the `model_update` objects the method will decrypt.
        :type key: `str`
        :param kwargs: Dictionary of arguments required to perform decryption \
        or partial decryption operations.
        :type kwargs: `dict`
        :return: The resulting decrypted plain text.
        :rtype: `list` of `np.ndarray`
        """
        raise NotImplementedError

    def avg_collected_ciphertext_response(self, lst_model_updates, **kwargs):
        """
        Receives a list of `model update`, where a `model_update` is of type
        `ModelUpdate`, using the encrypted values (indicating by the `key`)
        included in each `model_update`,
        it returns the decrypted resulting mean.

        :param: lst_model_updates. List of `model_update` of type
            `ModelUpdate` including the encrypted values (indicating by the `key`)
            to be averaged.
        :type lst_model_updates:  `list`

        :return: results of fused encrypted model update
        :rtype: `ModelUpdate`
        """
        fused_model_update = self.aggregate_collected_ciphertext_response(lst_model_updates)
        if fused_model_update.exist_key("weights"):
            avg_model_weights = fused_model_update.get("weights")
            for i in range(len(avg_model_weights)):
                avg_model_weights[i] = avg_model_weights[i] / len(lst_model_updates)
            avg_model_update = ModelUpdate(weights=avg_model_weights)
            return avg_model_update
        else:
            # fused model update is still in ciphertext format
            return fused_model_update

    @abc.abstractmethod
    def aggregate_collected_ciphertext_response(self, lst_model_updates, key="ct_weights", **kwargs):
        raise NotImplementedError


class Cipher(abc.ABC):
    """
    This class creates a crypto `Cipher` that encrypts and decrypts
    the message content exchanged during FL training.
    """

    def __init__(self, crypto_keys, precision, **kwargs):
        """
        Initializes a crypto `Cipher` for a crypto system with
        provided crypto keys, and set the `NEGATIVE_THRESHOLD` as the
        maximum size allowed by the system.

        :param crypto_keys: Provided crypto_keys, including public and \
        private key(s), for encryption and decryption.
        :type crypto_keys: `dict`
        :param precision: Provided precision for encoding and decoding, \
        default value is set to 6.
        :type precision: `int`
        :param kwargs: A dictionary contains additional arguments to \
        initialize a crypto `Cipher`.
        :type kwargs: `dict`
        """
        self.NEGATIVE_THRESHOLD = sys.maxsize
        self.keys = crypto_keys
        self.precision = precision

    def set_keys(self, keys):
        """
        Set the keys in use cases where the keys are not set during initialization
        of this object, e.g. for the keys generation and distribution protocol use cases.

        :param keys: The provided keys.
        :type keys: `dict`
        :return: None
        """
        raise NotImplementedError

    def encrypt(self, plaintext, **kwargs):
        raise NotImplementedError

    def decrypt(self, ciphertext, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def encrypt_weights(self, model_weights, **kwargs):
        """
        Encrypts the provided plain_text according to the crypto system setup,
         the encoding precision, and other encryption requirements defined
         inside the kwargs.

        :param model_weights: Plain text values to be encrypted.
        :type model_weights: `list` of `np.ndarray`
        :param kwargs: Encryption requirements
        :type kwargs: `dict`
        :return: Encrypted text
        :rtype: `list`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def decrypt_weights(self, model_weights_ciphertext, **kwargs):
        """
        Decrypts the provided cipher_text according to the crypto system
        setup, the encoding precision, and other decryption requirements
        defined inside the kwargs.

        :param model_weights_ciphertext: Cipher text to be decrypted.
        :type model_weights_ciphertext: `list`
        :param kwargs: Decryption requirements
        :type kwargs: `dict`
        :return: Decrypted text
        :rtype: `list`
        """
        raise NotImplementedError

    def _has_public_key(self):
        """
        Checks the existence of the public key(s).

        :return: True if the public key(s) exists otherwise returns False.
        :rtype: `boolean`
        """
        return "pk" in self.keys or "pp" in self.keys

    def _has_private_key(self):
        """
        Checks the existence of the private key(s).

        :return: True if the private key(s) exists otherwise returns False.
        :rtype: `boolean`
        """
        return "sk" in self.keys



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
import json, time
from ibm_watson_machine_learning.utils import StatusLogger, print_text_header_h1, print_text_header_h2
from ibm_watson_machine_learning.href_definitions import is_uid
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import  WMLClientError
from ibm_watson_machine_learning.metanames import SpacesPlatformMetaNames, SpacesPlatformMemberMetaNames
from ibm_watson_machine_learning.instance_new_plan import ServiceInstanceNewPlan
from ibm_watson_machine_learning.utils.deployment.errors import PromotionFailed

_DEFAULT_LIST_LENGTH = 50


class PlatformSpaces(WMLResource):
    """Store and manage spaces."""

    ConfigurationMetaNames = SpacesPlatformMetaNames()
    """MetaNames for spaces creation."""

    MemberMetaNames = SpacesPlatformMemberMetaNames()
    """MetaNames for space members creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._client = client

    def _get_resources(self, url, op_name, params=None):
        if params is not None and 'limit' in params.keys():
            if params[u'limit'] < 1:
                raise WMLClientError('Limit cannot be lower than 1.')
            elif params[u'limit'] > 1000:
                raise WMLClientError('Limit cannot be larger than 1000.')

        if len(params) > 0:
            response_get = requests.get(
                url,
                headers=self._client._get_headers(),
                params=params
            )

            return self._handle_response(200, op_name, response_get)
        else:

            resources = []

            while True:
                response_get = requests.get(url, headers=self._client._get_headers())

                result = self._handle_response(200, op_name, response_get)
                resources.extend(result['resources'])

                if 'next' not in result:
                    break
                else:
                    url = self._wml_credentials["url"]+result['next']['href']
                    if('start=invalid' in url):
                        break
            return {
                "resources": resources
            }

    def store(self, meta_props, background_mode=True):
        """Create a space. The instance associated with the space via COMPUTE will be used for billing purposes on
        cloud. Note that STORAGE and COMPUTE are applicable only for cloud.

        :param meta_props:  meta data of the space configuration. To see available meta names use:

            .. code-block:: python

                client.spaces.ConfigurationMetaNames.get()

        :type meta_props: dict

        :param background_mode: indicator if store() method will run in background (async) or (sync)
        :type background_mode: bool, optional

        :return: metadata of the stored space
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.spaces.ConfigurationMetaNames.NAME: 'my_space',
                client.spaces.ConfigurationMetaNames.DESCRIPTION: 'spaces',
                client.spaces.ConfigurationMetaNames.STORAGE: {"resource_crn": "provide crn of the COS storage"},
                client.spaces.ConfigurationMetaNames.COMPUTE: {"name": "test_instance",
                                                               "crn": "provide crn of the instance"}
            }
            spaces_details = client.spaces.store(meta_props=metadata)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        PlatformSpaces._validate_type(meta_props, u'meta_props', dict, True)

        if ('compute' in meta_props or 'storage' in meta_props) and self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("'STORAGE' and 'COMPUTE' meta props are not applicable on "
                                 "IBM Cloud Pak® for Data. If using any of these, remove and retry")

        if 'storage' not in meta_props and self._client.CLOUD_PLATFORM_SPACES:
            raise WMLClientError("'STORAGE' is mandatory for cloud")

        if 'compute' in meta_props and self._client.CLOUD_PLATFORM_SPACES:
            if 'name' not in meta_props[u'compute'] or 'crn' not in meta_props[u'compute']:
                raise WMLClientError("'name' and 'crn' is mandatory for 'COMPUTE'")
            temp_meta = meta_props[u'compute']
            temp_meta.update({'type': 'machine_learning'})

            meta_props[u'compute'] = temp_meta

        space_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client

        )

        if 'compute' in meta_props and self._client.CLOUD_PLATFORM_SPACES:
            payload_compute = []
            payload_compute.append(space_meta[u'compute'])
            space_meta[u'compute'] = payload_compute

        creation_response = requests.post(
            self._client.service_instance._href_definitions.get_platform_spaces_href(),
            headers=self._client._get_headers(),
            json=space_meta)

        spaces_details = self._handle_response(202, u'creating new spaces', creation_response)

        # Cloud Convergence: Set self._client.wml_credentials['instance_id'] to instance_id
        # during client.set.default_space since that's where space is associated with client
        # and also in client.set.default_project
        #
        if 'compute' in spaces_details['entity'].keys() and self._client.CLOUD_PLATFORM_SPACES:
            instance_id = spaces_details['entity']['compute'][0]['guid']
            self._client.wml_credentials[u'instance_id'] = instance_id
            self._client.service_instance = ServiceInstanceNewPlan(self._client)
            self._client.service_instance.details = self._client.service_instance.get_details()


        if background_mode:
            print("Space has been created. However some background setup activities might still be on-going. "
                  "Check for 'status' field in the response. It has to show 'active' before space can be used. "
                  "If it's not 'active', you can monitor the state with a call to spaces.get_details(space_id). "
                  "Alternatively, use background_mode=False when calling client.spaces.store().")
            return spaces_details

        else:
            # note: monitor space status
            space_id = self.get_id(spaces_details)
            print_text_header_h1(u'Synchronous space creation with id: \'{}\' started'.format(space_id))

            status = spaces_details['entity']['status'].get('state')

            with StatusLogger(status) as status_logger:
                while status not in ['failed', 'error', 'completed', 'canceled', 'active']:
                    time.sleep(10)
                    spaces_details = self.get_details(space_id)
                    status = spaces_details['entity']['status'].get('state')
                    status_logger.log_state(status)
            # --- end note

            if u'active' in status:
                print_text_header_h2(u'\nCreating space  \'{}\' finished successfully.'.format(space_id))
            else:
                raise WMLClientError(
                    f"Space {space_id} creation failed with status: {spaces_details['entity']['status']}")

            return spaces_details

    @staticmethod
    def get_id(space_details):
        """Get space_id from space details.

        :param space_details: metadata of the stored space
        :type space_details: dict

        :return: space ID
        :rtype: str

        **Example**

        .. code-block:: python

            space_details = client.spaces.store(meta_props)
            space_id = client.spaces.get_id(space_details)
        """

        PlatformSpaces._validate_type(space_details, u'space_details', object, True)

        return WMLResource._get_required_element_from_dict(space_details, u'space_details',
                                                           [u'metadata', u'id'])

    @staticmethod
    def get_uid(space_details):
        """Get Unique Id of the space.

         *Deprecated:* Use ``get_id(space_details)`` instead.

         :param space_details: metadata of the space
         :type space_details: dict

         :return: Unique Id of space
         :rtype: str

        **Example**

        .. code-block:: python

            space_details = client.spaces.store(meta_props)
            space_uid = client.spaces.get_uid(space_details)

        """
        PlatformSpaces._validate_type(space_details, u'space_details', object, True)

        return WMLResource._get_required_element_from_dict(space_details, u'space_details',
                                                           [u'metadata', u'id'])

    def delete(self, space_id):
        """Delete a stored space.

        :param space_id: space ID
        :type space_id: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.spaces.delete(space_id)
        """
        PlatformSpaces._validate_type(space_id, u'space_id', str, True)

        space_endpoint = self._client.service_instance._href_definitions.get_platform_space_href(space_id)

        response_delete = requests.delete(space_endpoint, headers=self._client._get_headers())

        response = self._handle_response(202, u'space deletion', response_delete, False)

        print('DELETED')

        return response

    def get_details(self, space_id=None, limit=None, asynchronous=False, get_all=False):
        """Get metadata of stored space(s).

        :param space_id: space ID
        :type space_id: str, optional
        :param limit: applicable when `space_id` is not provided, otherwise `limit` will be ignored
        :type limit: str, optional
        :param asynchronous: if `True`, it will work as a generator
        :type asynchronous: bool, optional
        :param get_all:  if `True`, it will get all entries in 'limited' chunks
        :type get_all: bool, optional

        :return: metadata of stored space(s)
        :rtype: dict

        **Example**

        .. code-block:: python

            space_details = client.spaces.get_details(space_uid)
            space_details = client.spaces.get_details(limit=100)
            space_details = client.spaces.get_details(limit=100, get_all=True)
            space_details = []
            for entry in client.spaces.get_details(limit=100, asynchronous=True, get_all=True):
                space_details.extend(entry)
        """
        PlatformSpaces._validate_type(space_id, u'space_id', str, False)

        href = self._client.service_instance._href_definitions.get_platform_space_href(space_id)

        if space_id is not None:
            response_get = requests.get(href, headers=self._client._get_headers())

            return self._handle_response(200, 'Get space', response_get)

        else:
            return self._get_with_or_without_limit(self._client.service_instance._href_definitions.get_platform_spaces_href(),
                                                   limit,
                                                   'spaces',
                                                   summary=False,
                                                   pre_defined=False,
                                                   skip_space_project_chk=True,
                                                   _async=asynchronous,
                                                   _all=get_all)

    def list(self, limit=None, member=None, roles=None, return_as_df=True):
        """Print stored spaces in a table format. If limit is set to None there will be only first 50 records shown.

        :param limit: limit number of fetched records
        :type limit: int, optional
        :param member: filters the result list to only include spaces where the user with a matching user id
            is a member
        :type member: str, optional
        :param roles: limit number of fetched records
        :type roles: str, optional
        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed spaces or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        **Example**

        .. code-block:: python

            client.spaces.list()
        """

        PlatformSpaces._validate_type(limit, u'limit', int, False)
        href = self._client.service_instance._href_definitions.get_platform_spaces_href()

        params = {}

        if limit is not None:
            params.update({'limit': limit})

        if limit is None:
            params.update({'limit': 50})

        if member is not None:
            params.update({'member': member})

        if roles is not None:
            params.update({'roles': roles})

        space_resources = self._get_resources(href, 'spaces', params)[u'resources']

        # space_resources = self._get_no_space_artifact_details(href, None, limit, 'spaces')[u'resources']

        space_values = [(m[u'metadata'][u'id'],
                         m[u'entity'][u'name'],
                         m[u'metadata'][u'created_at']) for m in space_resources]

        if limit is None:
            print("Note: 'limit' is not provided. Only first 50 records will be displayed if the number of records "
                  "exceed 50")

        table = self._list(space_values, [u'ID', u'NAME', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table


    def update(self, space_id, changes):
        """Updates existing space metadata. 'STORAGE' cannot be updated.
        STORAGE and COMPUTE are applicable only for cloud.

        :param space_id: ID of space which definition should be updated
        :type space_id: str
        :param changes: elements which should be changed, where keys are ConfigurationMetaNames
        :type changes: dict

        :return: metadata of updated space
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.spaces.ConfigurationMetaNames.NAME:"updated_space",
                client.spaces.ConfigurationMetaNames.COMPUTE: {"name": "test_instance",
                                                               "crn": "v1:staging:public:pm-20-dev:us-south:a/09796a1b4cddfcc9f7fe17824a68a0f8:f1026e4b-77cf-4703-843d-c9984eac7272::"
                }
            }
            space_details = client.spaces.update(space_id, changes=metadata)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        if ('compute' in changes or 'storage' in changes) and self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("'STORAGE' and 'COMPUTE' meta props are not applicable on"
                                 "IBM Cloud Pak® for Data. If using any of these, remove and retry")

        if 'storage' in changes:
            raise WMLClientError("STORAGE cannot be updated")

        self._validate_type(space_id, u'space_id', str, True)
        self._validate_type(changes, u'changes', dict, True)

        details = self.get_details(space_id)

        if 'compute' in changes and self._client.CLOUD_PLATFORM_SPACES:
            changes[u'compute'][u'type'] = 'machine_learning'

            payload_compute = []
            payload_compute.append(changes[u'compute'])
            changes[u'compute'] = payload_compute

        print("changes in update: ", changes)

        patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details['entity'], changes)

        print("patch payload: ", patch_payload)

        href = self._client.service_instance._href_definitions.get_platform_space_href(space_id)

        response = requests.patch(href, json=patch_payload, headers=self._client._get_headers())

        updated_details = self._handle_response(200, u'spaces patch', response)

        # Cloud Convergence
        if 'compute' in updated_details['entity'].keys() and self._client.CLOUD_PLATFORM_SPACES:
            instance_id = updated_details['entity']['compute'][0]['guid']
            self._client.wml_credentials[u'instance_id'] = instance_id
            self._client.service_instance = ServiceInstanceNewPlan(self._client)
            self._client.service_instance.details = self._client.service_instance.get_details()

        return updated_details


#######SUPPORT FOR SPACE MEMBERS

    def create_member(self, space_id, meta_props):
        """Create a member within a space.

        :param space_id: ID of space which definition should be updated
        :type space_id: str
        :param meta_props:  metadata of the member configuration. To see available meta names use:

            .. code-block:: python

                client.spaces.MemberMetaNames.get()

        :type meta_props: dict

        :return: metadata of the stored member
        :rtype: dict

        .. note::
            * `role` can be any one of the following: "viewer", "editor", "admin"
            * `type` can be any one of the following: "user", "service"
            * `id` can be either service-ID or IAM-userID

        **Examples**

        .. code-block:: python

            metadata = {
                client.spaces.MemberMetaNames.MEMBERS: [{"id":"IBMid-100000DK0B",
                                                         "type": "user",
                                                         "role": "admin" }]
            }
            members_details = client.spaces.create_member(space_id=space_id, meta_props=metadata)

        .. code-block:: python

            metadata = {
                client.spaces.MemberMetaNames.MEMBERS: [{"id":"iam-ServiceId-5a216e59-6592-43b9-8669-625d341aca71",
                                                         "type": "service",
                                                         "role": "admin" }]
            }
            members_details = client.spaces.create_member(space_id=space_id, meta_props=metadata)
        """
        self._validate_type(space_id, u'space_id', str, True)

        PlatformSpaces._validate_type(meta_props, u'meta_props', dict, True)

        meta = {}

        if 'members' in meta_props:
            meta = meta_props
        elif 'member' in meta_props:
            dictionary = meta_props['member']
            payload = []
            payload.append(dictionary)
            meta['members'] = payload

        space_meta = self.MemberMetaNames._generate_resource_metadata(
            meta,
            with_validation=True,
            client=self._client
        )

        creation_response = requests.post(
            self._client.service_instance._href_definitions.get_platform_spaces_members_href(space_id),
            headers=self._client._get_headers(),
            json=space_meta)

        # TODO: Change response code one they change it to 201
        members_details = self._handle_response(200, u'creating new members', creation_response)

        return members_details

    def get_member_details(self, space_id, member_id):
        """Get metadata of member associated with a space.

        :param space_id: ID of space which definition should be updated
        :type space_id: str
        :param member_id: member ID
        :type member_id: str

        :return: metadata of member of a space
        :rtype: dict

        **Example**

        .. code-block:: python

            member_details = client.spaces.get_member_details(space_uid,member_id)
        """
        PlatformSpaces._validate_type(space_id, u'space_id', str, True)

        PlatformSpaces._validate_type(member_id, u'member_id', str, True)

        href = self._client.service_instance._href_definitions.get_platform_spaces_member_href(space_id, member_id)

        response_get = requests.get(href, headers=self._client._get_headers())

        return self._handle_response(200, 'Get space member', response_get)

    def delete_member(self, space_id, member_id):
        """Delete a member associated with a space.

        :param space_id:  space UID
        :type space_id: str
        :param member_id:  member UID
        :type member_id: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.spaces.delete_member(space_id,member_id)
        """
        PlatformSpaces._validate_type(space_id, u'space_id', str, True)
        PlatformSpaces._validate_type(member_id, u'member_id', str, True)

        member_endpoint = self._client.service_instance._href_definitions.get_platform_spaces_member_href(space_id, member_id)

        response_delete = requests.delete(member_endpoint, headers=self._client._get_headers())

        print('DELETED')

        return self._handle_response(204, u'space member deletion', response_delete, False)

    def update_member(self, space_id, member_id, changes):
        """Updates existing member metadata.

        :param space_id: ID of space
        :type space_id: str
        :param member_id: ID of member that needs to be updated
        :type member_id: str
        :param changes: elements which should be changed, where keys are ConfigurationMetaNames
        :type changes: dict

        :return: metadata of updated member
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.spaces.MemberMetaNames.MEMBER: {"role": "editor"}
            }
            member_details = client.spaces.update_member(space_id, member_id, changes=metadata)
        """
        self._validate_type(space_id, u'space_id', str, True)
        self._validate_type(member_id, u'member_id', str, True)

        self._validate_type(changes, u'changes', dict, True)

        details = self.get_member_details(space_id, member_id)

        # The member record is a bit different than most other type of records we deal w.r.t patch
        # There is no encapsulating object for the fields. We need to be consistent with the way we
        # provide the meta in create/patch. When we give with .MEMBER, _generate_patch_payload
        # will generate with /member patch. So, separate logic for member patch inline here
        changes1 = changes['member']

        # Union of two dictionaries. The one in changes1 will override existent ones in current meta
        details.update(changes1)

        id_str = {}
        role_str = {}
        type_str = {}
        state_str = {}

        # if 'id' in details:
        #     id_str["op"] = "replace"
        #     id_str["path"] = "/id"
        #     id_str["value"] = details[u'id']
        if 'role' in details:
            role_str["op"] = "replace"
            role_str["path"] = "/role"
            role_str["value"] = details[u'role']
        # if 'type' in details:
        #     type_str["op"] = "replace"
        #     type_str["path"] = "/type"
        #     type_str["value"] = details[u'type']
        if 'state' in details:
            state_str["op"] = "replace"
            state_str["path"] = "/state"
            state_str["value"] = details[u'state']

        patch_payload = []

        # if id_str:
        #     patch_payload.append(id_str)
        if role_str:
            patch_payload.append(role_str)
        # if type_str:
        #     patch_payload.append(type_str)
        if state_str:
            patch_payload.append(state_str)

         # patch_payload = self.MemberMetaNames._generate_patch_payload(details, changes, with_validation=True)

        href = self._client.service_instance._href_definitions.get_platform_spaces_member_href(space_id,member_id)

        response = requests.patch(href, json=patch_payload, headers=self._client._get_headers())

        updated_details = self._handle_response(200, u'members patch', response)

        return updated_details

    def list_members(self, space_id, limit=None, identity_type=None, role=None, state=None, return_as_df=True):
        """Print stored members of a space in a table format.
        If limit is set to None there will be only first 50 records shown.

        :param space_id: ID of space
        :type space_id: str
        :param limit: limit number of fetched records
        :type limit: int, optional
        :param identity_type: filter the members by type
        :type identity_type: str, optional
        :param role: filter the members by role
        :type role: str, optional
        :param state: filter the members by state
        :type state: str, optional
        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed members or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        **Example**

        .. code-block:: python

            client.spaces.list_members(space_id)
        """
        self._validate_type(space_id, u'space_id', str, True)

        params = {}

        if limit is not None:
            params.update({'limit': limit})

        if limit is None:
            params.update({'limit': 50})

        if identity_type is not None:
            params.update({'type': identity_type})

        if role is not None:
            params.update({'role': role})

        if state is not None:
            params.update({'state': state})

        href = self._client.service_instance._href_definitions.get_platform_spaces_members_href(space_id)

        member_resources = self._get_resources(href, 'space members', params)[u'resources']

        # space_values = [(m[u'metadata'][u'id'],
        #                  m[u'entity'][u'id'],
        #                  m[u'entity'][u'type'],
        #                  m[u'entity'][u'role'],
        #                  m[u'entity'][u'state'],
        #                  m[u'metadata'][u'created_at']) for m in member_resources]


        # self._list(space_values, [u'ID', u'IDENTITY',
        #                           u'IDENTITY_TYPE',
        #                           u'ROLE',
        #                           u'STATE',
        #                           u'CREATED'], limit, _DEFAULT_LIST_LENGTH)

        space_values = [(m[u'id'],
                         m[u'type'],
                         m[u'role'],
                         m[u'state']) if 'state' in m else
                        (m[u'id'],
                         m[u'type'],
                         m[u'role'],
                         None) for m in member_resources]

        if limit is None:
            print("Note: 'limit' is not provided. Only first 50 records will be displayed if the number of records "
                  "exceed 50")

        table = self._list(space_values, [u'ID',
                                          u'TYPE',
                                          u'ROLE',
                                          u'STATE'], limit, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table

    def promote(self, asset_id: str, source_project_id: str, target_space_id: str, rev_id: str = None) -> str:
        """Promote asset from project to space.

        :param asset_id: stored asset
        :type asset_id: str

        :param source_project_id: source project, from which asset is promoted
        :type source_project_id: str

        :param target_space_id: target space, where asset is promoted
        :type target_space_id: str

        :param rev_id: revision ID of the promoted asset
        :type rev_id: str, optional

        :return: promoted asset id
        :rtype: str

        **Examples**

        .. code-block:: python

            promoted_asset_id = client.spaces.promote(asset_id, source_project_id=project_id, target_space_id=space_id)
            promoted_model_id = client.spaces.promote(model_id, source_project_id=project_id, target_space_id=space_id)
            promoted_function_id = client.spaces.promote(function_id, source_project_id=project_id, target_space_id=space_id)
            promoted_data_asset_id = client.spaces.promote(data_asset_id, source_project_id=project_id, target_space_id=space_id)
            promoted_connection_asset_id = client.spaces.promote(connection_id, source_project_id=project_id, target_space_id=space_id)
        """
        promote_payload = {"spaceId": target_space_id,
                           "projectId": source_project_id,
                           "assetDescription": "Asset promoted by ibm_wml client"}

        if rev_id:
            promote_payload['revisionId'] = rev_id

        promote_href = self._client.service_instance._href_definitions.promote_asset_href(asset_id)
        response = requests.post(
            promote_href,
            headers=self._client._get_headers(),
            json=promote_payload
        )
        promotion_details = self._client.repository._handle_response(200, f'promote asset', response)

        try:
            promoted_asset_id = promotion_details['promotedAsset']['asset_id']
        except KeyError as key_err:
            raise PromotionFailed(source_project_id, target_space_id, promotion_details, reason=key_err)

        return promoted_asset_id








#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import INSTANCE_DETAILS_TYPE
from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.foundation_models.utils.enums import ModelTypes, DecodingMethods
from ibm_watson_machine_learning.metanames import GenTextParamsMetaNames, GenTextReturnOptMetaNames


_DEFAULT_LIST_LENGTH = 50


class Model(WMLResource):
    """Instantiate the model interface.

    :param model_id: the type of model to use
    :type model_id: str

    :param params: parameters to use during generate requests
    :type params: dict, optional

    :param credentials: credentials to Watson Machine Learning instance
    :type credentials: dict

    :param project_id: ID of the Watson Studio project
    :type project_id: str, optional

    :param space_id: ID of the Watson Studio Space
    :type space_id: str, optional

    :param verify: user can pass as verify one of following:

        - the path to a CA_BUNDLE file
        - the path of directory with certificates of trusted CAs
        - `True` - default path to truststore will be taken
        - `False` - no verification will be made
    :type verify: bool or str, optional

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models import Model
        from ibm_watson_machine_learning.metanames import GenTextParamsMetaNames as GenParams
        from ibm_watson_machine_learning.foundation_models.utils.enums import ModelTypes, DecodingMethods

        flan_ul2_params = {
            GenParams.MAX_NEW_TOKENS: 25,
        }

        flan_ul2_model = Model(
            model_id=ModelTypes.FLAN_UL2,
            params=flan_ul2_params,
            credentials={
                "apikey": "...",
                "url": "https://us-south.ml.cloud.ibm.com"
            },
            project_id="12345"
            )
    """

    def __init__(self,
                 model_id: str,
                 credentials: dict,
                 params: dict = None,
                 project_id: str = None,
                 space_id: str = None,
                 verify=None) -> None:

        self.model_id = model_id
        if isinstance(self.model_id, ModelTypes):
            self.model_id = self.model_id.value
        self.params = params
        self.project_id = project_id
        self.space_id = space_id
        self._client = APIClient(credentials, verify=verify)

        Model._validate_type(params, u'params', dict, False)

        if not self._client.CLOUD_PLATFORM_SPACES:
            raise WMLClientError.__init__(self, f"Operation is unsupported for this release.")

        if space_id:
            self._client.set.default_space(space_id)
        else:
            self._client.set.default_project(project_id)

        WMLResource.__init__(self, __name__, self._client)

    def get_details(self, limit=50, asynchronous=False, get_all=False):
        """Get model's details

        :return: model's details
        :rtype: dict

        **Example**

        .. code-block:: python

            flan_ul2_model.get_details()

        """

        url = self._client.service_instance._href_definitions.get_fm_specifications_href(limit)
        models = self._get_artifact_details(base_url=url, uid=None, limit=limit, resource_name='foundation_models', _async=asynchronous, _all=get_all)['resources']

        return next((item for item in models if item['model_id'] == self.model_id), None)

    def generate(self, prompt, params=None):
        """Given a text prompt as inputs, and parameters the selected model (model_id)
        will generate a completion text as generated_text.

        :param params: meta props for text generation, use ``client.foundation_models.ConfigurationMetaNames.show()`` to view the list of MetaNames
        :type params: dict

        :param prompt: the prompt string
        :type prompt: str,


        :return: scoring result containing generated content
        :rtype: dict


        **Example**

        .. code-block:: python

            q = "What is 1 + 1?"
            ul2_response = flan_ul2_model.generate(prompts=q)
            print(ul2_response['results'][0]['generated_text'])

        """
        generate_text_url = self._client.service_instance._href_definitions.get_fm_generation_href('text?version=2022-08-01')
        Model._validate_type(params, u'params', dict, False)
        Model._validate_type(prompt, u'prompts', str, True)
        payload = {
            "model_id": self.model_id,
            "input": prompt,
        }

        if params:
            payload['parameters'] = params
        elif self.params:
            payload['parameters'] = self.params

        if 'parameters' in payload and GenTextParamsMetaNames.DECODING_METHOD in payload['parameters']:
            if isinstance(payload['parameters'][GenTextParamsMetaNames.DECODING_METHOD], DecodingMethods):
                payload['parameters'][GenTextParamsMetaNames.DECODING_METHOD] = payload['parameters'][GenTextParamsMetaNames.DECODING_METHOD].value

        if self._client.default_project_id:
            payload['project_id'] = self._client.default_project_id
        elif self._client.default_space_id:
            payload['space_id'] = self._client.default_space_id

        response_scoring = requests.post(
                url=generate_text_url,
                json=payload,
                headers=self._client._get_headers())

        return self._handle_response(200, u'generate', response_scoring)

    def generate_text(self, prompt, params=None):
        """Given a text prompt as inputs, and parameters the selected model (model_id)
        will generate a completion text as generated_text.

        :param params: meta props for text generation, use ``client.foundation_models.ConfigurationMetaNames.show()`` to view the list of MetaNames
        :type params: dict

        :param prompt: the prompt string
        :type prompt: str,


        :return: generated content
        :rtype: str


        **Example**

        .. code-block:: python

            q = "What is 1 + 1?"
            ul2_text = flan_ul2_model.generate_text(prompts=q)

        """

        return self.generate(prompt=prompt, params=params)['results'][0]['generated_text']



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.libs.repo.mlrepository import ModelArtifact


class ScikitModelArtifact(ModelArtifact):
    """
    Class representing model artifact.

    :param str uid: optional, uid which indicate that artifact already exists in repository service
    :param str name: optional, name of artifact
    :param MetaProps meta_props: optional, props used by other services
    """
    def __init__(self, uid, name, meta_props):
        super(ScikitModelArtifact, self).__init__(uid, name, meta_props)

    def pipeline_artifact(self):
        """
        Returns None. pipeline artifact for scikit model has not been implemented.

        :rtype: ModelArtifact
        """
        pass



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames, MetaProps, WmlLibrariesArtifact
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.libraries_artifact_reader import LibrariesArtifactReader
from ibm_watson_machine_learning.libs.repo.util.exceptions import MetaPropMissingError

class LibrariesArtifact(WmlLibrariesArtifact):
    """
    Class representing LibrariesArtifact artifact.

    :param str uid: optional, uid which indicate that artifact already exists in repository service
    :param str name: optional, name of artifact
    :param str description: optional, description of artifact
    :param str version: version of libraries
    :param str  platform of libraries
    """
    def __init__(self, library=None,  uid=None, name=None, meta_props=MetaProps({})):
        super(LibrariesArtifact, self).__init__(uid, name, meta_props)
        self.library = library
        self.uid = uid
        self.meta_props = meta_props
        if meta_props.prop(MetaNames.LIBRARIES.PATCH_INPUT) is None:
            if meta_props.prop(MetaNames.LIBRARIES.NAME) is None:
                raise MetaPropMissingError('Value specified for "meta_props" does not contain value for ''"MetaNames.LIBRARIES.NAME"')
            self.name = meta_props.get()[MetaNames.LIBRARIES.NAME]

            if meta_props.prop(MetaNames.LIBRARIES.VERSION) is None:
                raise MetaPropMissingError('Value specified for "meta_props" does not contain value for' '"MetaNames.LIBRARIES.VERSION"')

            if meta_props.prop(MetaNames.LIBRARIES.PLATFORM) is None:
                raise MetaPropMissingError('Value specified for "meta_props" does not contain value for '
                                       '"MetaNames.LIBRIES.PLATFORM"')


        #if meta_props.prop(MetaNames.LIBRARIES.NAME) is not None and not isinstance(meta_props.prop(MetaNames.LIBRARIES.NAME), str):
        #    raise ValueError('Invalid type for MetaNames.LIBRARIES.NAME: {}'.
        #                     format(meta_props.prop(MetaNames.LIBRARIES.NAME).__class__.__name__))
#
#        if meta_props.prop(MetaNames.LIBRARIES.VERSION) is not None and not isinstance(meta_props.prop(MetaNames.LIBRARIES.VERSION), str):
#            raise ValueError('Invalid type for MetaNames.LIBRARIES.VERSION: {}'.
#                             format(meta_props.prop(MetaNames.LIBRARIES.VERSION).__class__.__name__))
#

    def reader(self):
        """
        Returns reader used for getting library content.

        :return: reader for LibrariesArtifact.library
        :rtype: LibrariesArtifactReader
        """
        try:
            return self._reader
        except:
            self._reader = LibrariesArtifactReader(self.library)
            return self._reader

    def _copy(self, uid=None, meta_props=None):
        if uid is None:
            uid = self.uid

        if meta_props is None:
            meta_props = self.meta

        return LibrariesArtifact(
            self.library,
            uid=uid,
            meta_props=meta_props
        )



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import io
import os
import time
from typing import TYPE_CHECKING, Any, Dict, Union, List, Optional

import pandas as pd
from pandas import DataFrame, concat

from .base_deployment import BaseDeployment
from ..helpers import DataConnection, AssetLocation
from ..utils import StatusLogger, print_text_header_h1
from ..utils.autoai.connection import validate_source_data_connections, validate_deployment_output_connection
from ..utils.autoai.utils import init_cos_client, try_load_dataset, convert_dataframe_to_fields_values_payload
from ..utils.autoai.errors import NoneDataConnection
from ..utils.deployment.errors import BatchJobFailed, MissingScoringResults
from ..wml_client_error import WMLClientError
from ibm_watson_machine_learning.utils.autoai.enums import DataConnectionTypes

if TYPE_CHECKING:
    from sklearn.pipeline import Pipeline
    from pandas import DataFrame
    from numpy import ndarray
    from ..workspace import WorkSpace

__all__ = [
    "Batch"
]


class Batch(BaseDeployment):
    """The Batch Deployment class.
    With this class object you can manage any batch deployment.

    :param source_wml_credentials: credentials to Watson Machine Learning instance where training was performed
    :type source_wml_credentials: dict

    :param source_project_id: ID of the Watson Studio project where training was performed
    :type source_project_id: str, optional

    :param source_space_id: ID of the Watson Studio Space where training was performed
    :type source_space_id: str, optional

    :param target_wml_credentials: credentials to Watson Machine Learning instance where you want to deploy
    :type target_wml_credentials: dict

    :param target_project_id: ID of the Watson Studio project where you want to deploy
    :type target_project_id: str, optional

    :param target_space_id: ID of the Watson Studio Space where you want to deploy
    :type target_space_id: str, optional
    """

    def __init__(self,
                 source_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 source_project_id: str = None,
                 source_space_id: str = None,
                 target_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 target_project_id: str = None,
                 target_space_id: str = None,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None):

        super().__init__(
            deployment_type='batch',
            source_wml_credentials=source_wml_credentials,
            source_project_id=source_project_id,
            source_space_id=source_space_id,
            target_wml_credentials=target_wml_credentials,
            target_project_id=target_project_id,
            target_space_id=target_space_id,
            wml_credentials=wml_credentials,
            project_id=project_id,
            space_id=space_id
        )

        self.name = None
        self.id = None
        self.asset_id = None

    def __repr__(self):
        return f"name: {self.name}, id: {self.id}, asset_id: {self.asset_id}"

    def __str__(self):
        return f"name: {self.name}, id: {self.id}, asset_id: {self.asset_id}"

    def score(self, **kwargs):
        raise NotImplementedError("Batch deployment supports only job runs.")

    def create(self,
               model: str,
               deployment_name: str,
               metadata: Optional[Dict] = None,
               training_data: Optional[Union['DataFrame', 'ndarray']] = None,
               training_target: Optional[Union['DataFrame', 'ndarray']] = None,
               experiment_run_id: Optional[str] = None) -> None:
        """Create deployment from a model.

        :param model: AutoAI model name
        :type model: str

        :param deployment_name: name of the deployment
        :type deployment_name: str

        :param training_data: training data for the model
        :type training_data: pandas.DataFrame or numpy.ndarray, optional

        :param training_target: target/label data for the model
        :type training_target: pandas.DataFrame or numpy.ndarray, optional

        :param metadata: model meta properties
        :type metadata: dict, optional

        :param experiment_run_id: ID of a training/experiment (only applicable for AutoAI deployments)
        :type experiment_run_id: str, optional

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.deployment import Batch

            deployment = Batch(
                   wml_credentials={
                         "apikey": "...",
                         "iam_apikey_description": "...",
                         "iam_apikey_name": "...",
                         "iam_role_crn": "...",
                         "iam_serviceid_crn": "...",
                         "instance_id": "...",
                         "url": "https://us-south.ml.cloud.ibm.com"
                       },
                    project_id="...",
                    space_id="...")

            deployment.create(
                   experiment_run_id="...",
                   model=model,
                   deployment_name='My new deployment'
               )
        """
        return super().create(model=model,
                              deployment_name=deployment_name,
                              metadata=metadata,
                              training_data=training_data,
                              training_target=training_target,
                              experiment_run_id=experiment_run_id,
                              deployment_type='batch')

    @BaseDeployment._project_to_space_to_project
    def get_params(self) -> Dict:
        """Get deployment parameters."""
        return super().get_params()

    @BaseDeployment._project_to_space_to_project
    def run_job(self,
                payload: Union[DataFrame, List[DataConnection], Dict[str, DataFrame], Dict[str, DataConnection]] = pd.DataFrame(),
                output_data_reference: 'DataConnection' = None,
                transaction_id: str = None,
                background_mode: 'bool' = True) -> Union[Dict, Dict[str, List], DataConnection]:
        """Batch scoring job on WML. Payload or Payload data reference is required.
        It is passed to the WML where model have been deployed.

        :param payload: DataFrame that contains data to test the model or data storage connection details
            that inform the model where payload data is stored
        :type payload: pandas.DataFrame or List[DataConnection] or Dict

        :param output_data_reference: DataConnection to the output COS for storing predictions,
             required only when DataConnections are used as a payload
        :type output_data_reference: DataConnection, optional

        :param transaction_id: can be used to indicate under which id the records will be saved into payload table
            in IBM OpenScale
        :type transaction_id: str, optional

        :param background_mode: indicator if score() method will run in background (async) or (sync)
        :type background_mode: bool, optional

        :return: scoring job details
        :rtype: dict

        **Examples**

        .. code-block:: python

            score_details = batch_service.run_job(payload=test_data)
            print(score_details['entity']['scoring'])

            # Result:
            # {'input_data': [{'fields': ['sepal_length',
            #               'sepal_width',
            #               'petal_length',
            #               'petal_width'],
            #              'values': [[4.9, 3.0, 1.4, 0.2]]}],
            # 'predictions': [{'fields': ['prediction', 'probability'],
            #               'values': [['setosa',
            #                 [0.9999320742502246,
            #                  5.1519823540224506e-05,
            #                  1.6405926235405522e-05]]]}]

            payload_reference = DataConnection(location=DSLocation(asset_id=asset_id))
            score_details = batch_service.run_job(payload=payload_reference, output_data_filename = "scoring_output.csv")
            score_details = batch_service.run_job(payload={'observations': payload_reference})
            score_details = batch_service.run_job(payload=[payload_reference])
            score_details = batch_service.run_job(payload={'observations': payload_reference, 'supporting_features': supporting_features_reference})  # supporting features time series forecasting sceanrio
        """

        if isinstance(payload, dict):
            observations = payload.get('observations', pd.DataFrame())
            supporting_features = payload.get('supporting_features')

            if isinstance(observations, DataFrame) and \
                    (isinstance(supporting_features, DataFrame) or supporting_features is None):
                observations_payload = convert_dataframe_to_fields_values_payload(observations, return_values_only=True)
                observations_payload['id'] = 'observations'
                input_data = [observations_payload]

                if supporting_features is not None:
                    supporting_features_payload = convert_dataframe_to_fields_values_payload(supporting_features, return_values_only=True)
                    supporting_features_payload['id'] = 'supporting_features'
                    input_data.append(supporting_features_payload)

                scoring_payload = {
                    self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA: input_data
                }

            elif isinstance(observations, DataConnection) and \
                    (isinstance(supporting_features, DataConnection) or supporting_features is None):

                observations.id = 'observations'
                input_data = [observations]
                if supporting_features is not None:
                    supporting_features.id = 'supporting_features'
                    input_data.append(supporting_features)

                for data_conn in input_data:
                    if hasattr(data_conn, 'location') and isinstance(data_conn.location, AssetLocation):
                        data_conn.location.wml_client = self._target_workspace.wml_client

                input_data = validate_source_data_connections(source_data_connections=input_data,
                                                              workspace=self._target_workspace,
                                                              deployment=True)
                input_data = [data_connection._to_dict() for data_connection in input_data]

                if output_data_reference is None:
                    raise ValueError("\"output_data_reference\" should be provided.")

                if isinstance(output_data_reference, DataConnection):

                    # wml_client sets correct href for Data Assets
                    if hasattr(output_data_reference, 'location') and isinstance(output_data_reference.location,
                                                                                 AssetLocation):
                        output_data_reference.location.wml_client = self._target_workspace.wml_client

                    output_data_reference = validate_deployment_output_connection(
                        results_data_connection=output_data_reference,
                        workspace=self._target_workspace,
                        source_data_connections=input_data)
                    output_data_reference = output_data_reference._to_dict()

                scoring_payload = {
                    self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA_REFERENCES: input_data,
                    self._target_workspace.wml_client.deployments.ScoringMetaNames.OUTPUT_DATA_REFERENCE:
                        output_data_reference}

            else:
                raise TypeError('Missing data observations in payload or observations '
                                 'or supporting_features are not pandas.DataFrames.')
        # note: support for DataFrame payload
        elif isinstance(payload, DataFrame):
            scoring_payload = {
                self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA: [{'values': payload}]
            }
        # note: support for DataConnections and dictionaries payload
        elif isinstance(payload, list):
            if isinstance(payload[0], DataConnection):
                if None in payload:
                    raise NoneDataConnection('payload')

                # wml_client sets correct href for Data Assets
                for data_conn in payload:
                    if hasattr(data_conn, 'location') and isinstance(data_conn.location, AssetLocation):
                        data_conn.location.wml_client = self._target_workspace.wml_client

                payload = [new_conn for conn in payload for new_conn in conn._subdivide_connection()]
                payload = validate_source_data_connections(source_data_connections=payload,
                                                           workspace=self._target_workspace,
                                                           deployment=True)
                payload = [data_connection._to_dict() for data_connection in payload]
            elif isinstance(payload[0], dict):
                pass
            else:
                raise ValueError(f"Current payload type: list of {type(payload[0])} is not supported.")

            if output_data_reference is None:
                raise ValueError("\"output_data_reference\" should be provided.")

            if isinstance(output_data_reference, DataConnection):

                # wml_client sets correct href for Data Assets
                if hasattr(output_data_reference, 'location') and isinstance(output_data_reference.location, AssetLocation):
                    output_data_reference.location.wml_client = self._target_workspace.wml_client

                output_data_reference = validate_deployment_output_connection(
                    results_data_connection=output_data_reference,
                    workspace=self._target_workspace,
                    source_data_connections=payload)
                output_data_reference = output_data_reference._to_dict()

            scoring_payload = {
                self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA_REFERENCES: payload,
                self._target_workspace.wml_client.deployments.ScoringMetaNames.OUTPUT_DATA_REFERENCE:
                    output_data_reference}

        else:
            raise ValueError(
                f"Incorrect payload type. Required: DataFrame or List[DataConnection], Passed: {type(payload)}")

        scoring_payload['hybrid_pipeline_hardware_specs'] = [
            {
                'node_runtime_id': 'auto_ai.kb',
                'hardware_spec': {
                    'name': 'M'
                }
            }
        ]

        if self._obm:
            scoring_payload['hybrid_pipeline_hardware_specs'].insert(
                0,
                {
                    "node_runtime_id": "auto_ai.obm",
                    "hardware_spec": {
                        "name": "M-Spark",
                        "num_nodes": 2
                    }
                }
            )

        job_details = self._target_workspace.wml_client.deployments.create_job(self.id,
                                                                               scoring_payload, _asset_id=self.asset_id)

        if background_mode:
            return job_details

        else:
            # note: monitor scoring job
            job_id = self._target_workspace.wml_client.deployments.get_job_uid(job_details)
            print_text_header_h1(u'Synchronous scoring for id: \'{}\' started'.format(job_id))

            status = self.get_job_status(job_id)['state']

            with StatusLogger(status) as status_logger:
                while status not in ['failed', 'error', 'completed', 'canceled']:
                    time.sleep(10)
                    status = self.get_job_status(job_id)['state']
                    status_logger.log_state(status)
            # --- end note

            if u'completed' in status:
                print(u'\nScoring job  \'{}\' finished successfully.'.format(job_id))
            else:
                raise BatchJobFailed(job_id, f"Scoring job failed with status: {self.get_job_status(job_id)}")

            return self.get_job_params(job_id)

    @BaseDeployment._project_to_space_to_project
    def rerun_job(self,
                  scoring_job_id: str,
                  background_mode: bool = True) -> Union[dict, 'DataFrame', 'DataConnection']:
        """Rerun scoring job with the same parameters as job described by `scoring_job_id`.

        :param scoring_job_id: Id described scoring job
        :type scoring_job_id: str

        :param background_mode: indicator if score_rerun() method will run in background (async) or (sync)
        :type background_mode: bool, optional

        :return: scoring job details
        :rtype: dict

        **Example**

        .. code-block:: python

            scoring_details = deployment.score_rerun(scoring_job_id)
        """
        scoring_params = self.get_job_params(scoring_job_id)['entity']['scoring']
        input_data_references = self._target_workspace.wml_client.deployments.ScoringMetaNames.INPUT_DATA_REFERENCES
        output_data_reference = self._target_workspace.wml_client.deployments.ScoringMetaNames.OUTPUT_DATA_REFERENCE

        if input_data_references in scoring_params:
            payload_ref = [input_ref for input_ref in scoring_params[input_data_references]]

            if 'href' in scoring_params[output_data_reference]['location']:
                del scoring_params[output_data_reference]['location']['href']

            return self.run_job(payload=payload_ref, output_data_reference=scoring_params['output_data_reference'],
                                background_mode=background_mode)
        else:
            raise NotImplementedError("'rerun_job' method supports only jobs with "
                                      "payload passed as a list of DataConnections. If you want to rerun job "
                                      "with payload passed directly, please use 'run_job' one more time.")

    @BaseDeployment._project_to_space_to_project
    def delete(self, deployment_id: str = None) -> None:
        """Delete deployment on WML.

        :param deployment_id: ID of the deployment to delete, if empty, current deployment will be deleted
        :type deployment_id: str, optional

        **Example**

        .. code-block:: python

            deployment = Batch(workspace=...)
            # Delete current deployment
            deployment.delete()
            # Or delete a specific deployment
            deployment.delete(deployment_id='...')
        """
        super().delete(deployment_id=deployment_id, deployment_type='batch')

    @BaseDeployment._project_to_space_to_project
    def list(self, limit=None) -> 'DataFrame':
        """List WML deployments.

        :param limit: set the limit of how many deployments to list,
            default is `None` (all deployments should be fetched)
        :type limit: int, optional

        :return: Pandas DataFrame with information about deployments
        :rtype: pandas.DataFrame

        **Example**

        .. code-block:: python

            deployment = Batch(workspace=...)
            deployments_list = deployment.list()
            print(deployments_list)

            # Result:
            #                  created_at  ...  status
            # 0  2020-03-06T10:50:49.401Z  ...   ready
            # 1  2020-03-06T13:16:09.789Z  ...   ready
            # 4  2020-03-11T14:46:36.035Z  ...  failed
            # 3  2020-03-11T14:49:55.052Z  ...  failed
            # 2  2020-03-11T15:13:53.708Z  ...   ready
        """
        return super().list(limit=limit, deployment_type='batch')

    @BaseDeployment._project_to_space_to_project
    def get(self, deployment_id: str) -> None:
        """Get WML deployment.

        :param deployment_id: ID of the deployment to work with
        :type deployment_id: str

        **Example**

        .. code-block:: python

            deployment = Batch(workspace=...)
            deployment.get(deployment_id="...")
        """
        super().get(deployment_id=deployment_id, deployment_type='batch')

    @BaseDeployment._project_to_space_to_project
    def get_job_params(self, scoring_job_id: str = None) -> Dict:
        """Get batch deployment job parameters.

        :param scoring_job_id: Id of scoring job
        :type scoring_job_id: str

        :return: parameters of the scoring job
        :rtype: dict
        """
        return self._target_workspace.wml_client.deployments.get_job_details(scoring_job_id)

    @BaseDeployment._project_to_space_to_project
    def get_job_status(self, scoring_job_id: str) -> Dict:
        """Get status of scoring job.

        :param scoring_job_id: Id of scoring job
        :type scoring_job_id: str

        :return: dictionary with state of scoring job (one of: [completed, failed, starting, queued])
            and additional details if they exist
        :rtype: dict
        """
        return self._target_workspace.wml_client.deployments.get_job_status(scoring_job_id)

    @BaseDeployment._project_to_space_to_project
    def get_job_result(self, scoring_job_id: str) -> 'DataFrame':
        """Get batch deployment results of job with id `scoring_job_id`.

        :param scoring_job_id: Id of scoring job which results will be returned
        :type scoring_job_id: str

        :return: result
        :rtype: pandas.DataFrame

        :raises MissingScoringResults: in case of incompleted or failed job
            `MissingScoringResults` scoring exception is raised
        """
        scoring_params = self.get_job_params(scoring_job_id)['entity']['scoring']
        if scoring_params['status']['state'] == 'completed':
            if 'predictions' in scoring_params:
                data = DataFrame(scoring_params['predictions'][0]['values'], columns=scoring_params['predictions'][0]['fields'])
                return data
            else:
                conn = DataConnection._from_dict(scoring_params['output_data_reference'])
                conn._wml_client = self._target_workspace.wml_client

                return conn.read(raw=True) # if in future output may be excel file or with custom separator, here it should be recognized
        else:
            raise MissingScoringResults(scoring_job_id, reason="Scoring is not completed.")

    @BaseDeployment._project_to_space_to_project
    def get_job_id(self, batch_scoring_details):
        """Get id from batch scoring details."""
        return self._target_workspace.wml_client.deployments.get_job_uid(batch_scoring_details)

    @BaseDeployment._project_to_space_to_project
    def list_jobs(self):
        """Returns pandas DataFrame with list of deployment jobs"""

        resources = self._target_workspace.wml_client.deployments.get_job_details()['resources']
        columns = [u'job id', u'state', u'creted', u'deployment id']
        values = []
        for scoring_details in resources:
            if 'scoring' in scoring_details['entity']:
                state = scoring_details['entity']['scoring']['status']['state']
                score_values = (scoring_details[u'metadata'][u'id'], state,
                                scoring_details[u'metadata'][u'created_at'],
                                scoring_details['entity']['deployment']['id'])
                if self.id:
                    if self.id == scoring_details['entity']['deployment']['id']:
                        values.append(score_values)
                else:
                    values.append(score_values)

        return DataFrame(values, columns=columns)

    @BaseDeployment._project_to_space_to_project
    def _deploy(self,
                pipeline_model: 'Pipeline',
                deployment_name: str,
                meta_props: Dict,
                serving_name=None, # Not used, but added to match unified parameters for _deploy
                result_client=None,
                hardware_spec=None) -> Dict: # Not used, but added to match unified parameters for _deploy
        """Deploy model into WML.

        :param pipeline_model: model of the pipeline to deploy
        :type pipeline_model: Pipeline or str

        :param deployment_name: name of the deployment
        :type deployment_name: str

        :param meta_props: model meta properties
        :type meta_props: dict

        :param result_client: tuple with Result DataConnection object and initialized COS client
        :type result_client: tuple[DataConnection, resource]
        """
        deployment_details = {}
        asset_uid = self._publish_model(pipeline_model=pipeline_model,
                                        meta_props=meta_props)

        self.asset_id = asset_uid

        deployment_props = {
            self._target_workspace.wml_client.deployments.ConfigurationMetaNames.NAME: deployment_name,
            self._target_workspace.wml_client.deployments.ConfigurationMetaNames.BATCH: {}
        }

        deployment_props[self._target_workspace.wml_client.deployments.ConfigurationMetaNames.ASSET] = {
            "id": asset_uid
        }

        deployment_props[
            self._target_workspace.wml_client.deployments.ConfigurationMetaNames.HYBRID_PIPELINE_HARDWARE_SPECS] = [
            {
                'node_runtime_id': 'auto_ai.kb',
                'hardware_spec': {
                    'name': 'M'
                }
            }
        ]

        if self._obm:
            deployment_props[
                self._target_workspace.wml_client.deployments.ConfigurationMetaNames.HYBRID_PIPELINE_HARDWARE_SPECS
            ].insert(0, {
                "node_runtime_id": "auto_ai.obm",
                "hardware_spec": {
                    "name": "M-Spark",
                    "num_nodes": 2
                }
            })

        print("Deploying model {} using V4 client.".format(asset_uid))
        try:
            deployment_details = self._target_workspace.wml_client.deployments.create(
                artifact_uid=asset_uid,
                meta_props=deployment_props)

            self.deployment_id = self._target_workspace.wml_client.deployments.get_uid(deployment_details)

        except WMLClientError as e:
            raise e

        return deployment_details



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from warnings import warn

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.wml_client_error import MissingArgument, WrongEnvironmentVersion
from ..utils.autoai.enums import TShirtSize
from ..utils.autoai.errors import TShirtSizeNotSupported, SetIDFailed
from ..utils.autoai.utils import is_ipython


class WorkSpace:
    """WorkSpace class for WML authentication and project/space manipulation."""

    def __init__(self, wml_credentials: dict, project_id: str = None, space_id: str = None, verify=None) -> None:
        """
        :param wml_credentials: credentials to Watson Machine Learning instance
        :type wml_credentials: dict

        :param project_id: ID of the Watson Studio project
        :type project_id: str, optional

        :param space_id: ID of the Watson Studio Space
        :type space_id: str, optional

        **Example**

        .. code-block: python

            from ibm_watson_machine_learning.workspace import WorkSpace

            ws = WorkSpace(
                wml_credentials={
                    "apikey": "...",
                    "iam_apikey_description": "...",
                    "iam_apikey_name": "...",
                    "iam_role_crn": "...",
                    "iam_serviceid_crn": "...",
                    "instance_id": "...",
                    "url": "https://us-south.ml.cloud.ibm.com"
                },
                project_id="...",
                space_id="...")
        """
        self.wml_credentials = wml_credentials.copy()
        self.project_id = project_id
        self.space_id = space_id
        self.WMLS = False

        # note: base client changed versioning for WMLS
        if wml_credentials.get('instance_id', '').lower() == 'wml_local':
            supported_versions = ('2.0')
            if wml_credentials[u'version'] not in supported_versions:
                raise WrongEnvironmentVersion(wml_credentials[u'version'], wml_credentials[u'instance_id'].lower(),
                                              supported_versions)

            else:
                self.wml_credentials['version'] = '3.0.0'
        # --- end note

        self.wml_client = APIClient(wml_credentials=self.wml_credentials, verify=verify)

        if not wml_credentials.get('instance_id'):
            if self.space_id is not None:
                self.wml_client.set.default_space(self.space_id)

            elif self.project_id is not None:
                self.wml_client.set.default_project(self.project_id)

            else:
                raise SetIDFailed(
                    f'project_id and space_id',
                    reason=f'project_id and space_id cannot be None at the same time.')

        elif wml_credentials.get('instance_id', '').lower() == 'wml_local':
            print("Connecting to WML Server...")
            if self.space_id is None:
                raise MissingArgument(
                    'space_id',
                    reason="These credentials are from WML Server environment, "
                           "please specify the \"space_id\"")
            else:
                outcome = self.wml_client.set.default_space(self.space_id)
                if outcome == 'FAILURE':
                    raise SetIDFailed(
                        'space_id',
                        reason=f"This space_id: {self.space_id} cannot be found in current environment.")
                self.WMLS = True

        elif wml_credentials.get('instance_id', '').lower() in ('icp', 'openshift'):
            if self.project_id is None and self.space_id is None:
                raise MissingArgument(
                    'project_id',
                    reason="These credentials are from CP4D environment, "
                           "please specify \"project_id\"")
            else:
                if self.project_id is not None:
                    outcome = self.wml_client.set.default_project(self.project_id)

                else:
                    outcome = self.wml_client.set.default_space(self.space_id)

                if outcome == 'FAILURE':
                    raise SetIDFailed(
                        f'{"project_id" if self.project_id is not None else "space_id"}',
                        reason=f'This {"project_id" if self.project_id is not None else "space_id"}: '
                               f'{self.project_id if self.project_id is not None else self.space_id} '
                               f'cannot be found in current environment.')

            supported_versions = ('3.0.0', '3.0.1', '3.5', '4.0', '4.5', '4.6', '4.7')
            if wml_credentials[u'version'] not in supported_versions:
                raise WrongEnvironmentVersion(wml_credentials[u'version'], wml_credentials[u'instance_id'].lower(),
                                              supported_versions)

    def __str__(self):
        return f'wml_credentials: {self.wml_credentials} project_id: {self.project_id} space_id = {self.space_id}'

    def __repr__(self):
        return self.__str__()

    def restrict_pod_size(self, t_shirt_size: 'TShirtSize') -> 'TShirtSize':
        """Check t_shirt_size for AutoAI POD. Restrict sizes per environment.

        :param t_shirt_size: TShirt size to be validated and restricted
        :type t_shirt_size: TShirtSize

        :return: validated and restricted TShirt size
        :rtype: TShirtSize
        """
        # note: for testing purposes
        if self.wml_credentials.get('development', False):
            return t_shirt_size
        # --- end note

        default_cloud = TShirtSize.L
        default_cp4d = TShirtSize.M
        default_server = TShirtSize.M

        supported_cp4d = (TShirtSize.S, TShirtSize.M, TShirtSize.L, TShirtSize.XL)
        supported_server = (TShirtSize.S, TShirtSize.M, TShirtSize.L)

        # note: check CP4D and Server pod sizes
        if self.wml_client.ICP:
            if self.WMLS and t_shirt_size not in supported_server:
                message = (f"This t-shirt-size: \"{t_shirt_size}\" is not supported in WMLS. "
                           f"Supported sizes: {supported_server} "
                           f"Continuing work with default size {default_server}")

                warn(message=message)
                if is_ipython():
                    print(message)

                return default_server

            elif t_shirt_size not in supported_cp4d:
                message = (f"This t-shirt-size: \"{t_shirt_size}\" is not supported in CP4D. "
                           f"Supported sizes: {supported_cp4d} "
                           f"Continuing work with default size {default_cp4d}")

                warn(message=message)
                if is_ipython():
                    print(message)

                return default_cp4d

            else:
                return t_shirt_size

        else:
            # note: allow every size in test envs
            if 'test' in self.wml_credentials.get('url', ''):
                return t_shirt_size

            else:
                # note: raise an error for cloud if pod size is different, other just return
                if t_shirt_size != default_cloud:
                    raise TShirtSizeNotSupported(
                        t_shirt_size,
                        reason=f"This t-shirt size is not supported. Please use the default one: {default_cloud}")

                else:
                    return default_cloud
                # --- end note
